# $Id: frepp.mk,v 18.0 2010/03/03 00:02:17 fms Exp $
# frepp.mk: a Makefile approach to frepp
# using three-level storage model

#verbose flag: comment out this line to turn off verbose
verbose :=# -v

#use csh, with no user .cshrc, MUCH SLOWER without -f!
SHELL := /bin/csh -f

#site configuration, commands used by this Makefile
include site-config.mk
#experiment configuration, imported from FRE XML
include expt-config.mk

# list of available history files
HISTORIES := $(shell ls $(ARCHROOT)/history | grep .nc.cpio)
HISTDATES := $(patsubst %.nc.cpio,%,$(HISTORIES))
HISTYEARS := $(shell ls $(ARCHROOT)/history | grep .nc.cpio | cut -c1-4 |sort -u)
HISTSTART := $(firstword $(HISTYEARS))
HISTEND := $(word $(words $(HISTYEARS)), $(HISTYEARS))
#CHUNKSTART: year to start chunking (must have histstart<=chunkstart<=procstart)
CHUNKSTART ?= $(HISTSTART)
#PROCSTART: year to start processing (from FRE), defaults to HISTSTART
PROCSTART ?= $(HISTSTART)
#PROCEND: year to end processing (from FRE), defaults to HISTEND
PROCEND ?= $(HISTEND)
PROCYEARS = $(wordlist $(call nsince,$(PROCSTART),$(HISTSTART)),$(call nsince,$(PROCEND),$(HISTSTART)),$(HISTYEARS))

ifdef $(GRIDSPEC_SOURCE)
GRIDSPEC_ROOT := $(basename $(notdir $(GRIDSPEC_SOURCE)))
GRIDSPEC := $(WORKROOT)/$(GRIDSPEC_ROOT)
ifeq ($(suffix $(GRIDSPEC_SOURCE)),.cpio)
GRIDSPEC_FILES := atmos_mosaic.nc $(foreach tile, $(atmos_TILES), horizontal_grid$(tile).nc)
GRIDSPEC_PATHS := $(addprefix $(GRIDSPEC)/, $(GRIDSPEC_FILES))
else
#write clause for no-mosaic case
endif
endif

# functions
#remove PTMPROOT/ from a ptmp path
rpath = $(subst $(PTMPROOT)/,,$(1))
#convert a ptmp path to a work path
wpath = $(subst $(PTMPROOT),$(WORKROOT),$(1))
#find history file containing a given date
container = $(shell $(CONTAINER) $(1) $(HISTDATES))
#expr modulo is negative for negative numbers
modulo = $(shell expr $(1) % $(2))
#ysince is b-a+1, but eliminates answers <=0 (otherwise wordlist fails...)
# ysince = $(shell expr $(1) - $(2) + 1 |sed s/-// |sed s/^0/1/)
ysince = $(shell printf %4.4u `expr $(1) - $(2) + 1`)
nsince = $(shell expr $(1) - $(2) + 1)
#list of years from $(1) to $(2)
ylist = $(wordlist $(call nsince,$(1),$(HISTSTART)), $(call nsince,$(2),$(HISTSTART)), $(HISTYEARS))

#### section to define chunks

# first make list of all possible chunks and their beginning and ending year

define make_chunk_list
chunk_list += $(1)
endef
$(foreach comp, $(COMPONENTS), \
$(foreach freq, $($(comp)_TS_FREQS), \
$(eval $(call make_chunk_list, $($(comp)_TS_$(freq)_CHUNKS)))))
$(foreach comp, $(COMPONENTS), \
$(foreach src, $($(comp)_AV_SOURCES), \
$(eval $(call make_chunk_list, $($(comp)_TS_$(src)_INTERVALS)))))
CHUNKS := $(sort $(chunk_list))

define define_chunk
#1=year 2=chunk
ifeq ($(call modulo, $(call nsince,$(1),$(CHUNKSTART)),$(2)), 0)
end$(2) += $(1)
beg$(2) += $(call ysince,$(1),$(2))
endif
endef
$(foreach year, $(PROCYEARS), \
$(foreach chunk, $(CHUNKS), \
$(eval $(call define_chunk,$(year),$(chunk)))))

define debug_chunk
# 1=chunk length
phony += $(debug_$(1)_year_chunks)
debug_$(1)_year_chunks:
	@echo $(1)-year chunks begin: $(beg$(1))
	@echo $(1)-year chunks end:   $(end$(1))
endef
$(foreach chunk, $(CHUNKS), $(eval $(call debug_chunk,$(chunk))))

define debug_chunk_years
# 1=ending year 2=chunk length
phony += $(debug_$(2)_year_chunk_ending_$(1)
debug_$(2)_year_chunk_ending_$(1):
	@echo $(2)-year chunk ending at $(1): $(call ylist,$(call ysince,$(1),$(CHUNKSTART)),$(2))
endef
$(foreach chunk, $(CHUNKS), \
$(foreach year, $(end$(chunk)), \
$(eval $(call debug_chunk_years,$(year),$(chunk)))))

BDATE_3hr := 010101
EDATE_3hr := 123123
BDATE_daily := 0101
EDATE_daily := 1231
BDATE_monthly := 01
EDATE_monthly := 12
define define_ts_ranges
$(1)_TS_RANGES             += $(call ysince,$(4),$(3))$(BDATE_$(2))-$(4)$(EDATE_$(2))
$(1)_TS_$(2)_RANGES        += $(call ysince,$(4),$(3))$(BDATE_$(2))-$(4)$(EDATE_$(2))
$(1)_TS_$(2)_$(3)yr_RANGES += $(call ysince,$(4),$(3))$(BDATE_$(2))-$(4)$(EDATE_$(2))
endef
$(foreach comp, $(COMPONENTS), \
$(foreach freq, $($(comp)_TS_FREQS), \
$(foreach chunk, $($(comp)_TS_$(freq)_CHUNKS), \
$(foreach year, $(end$(chunk)), \
$(eval $(call define_ts_ranges,$(comp),$(freq),$(chunk),$(year)))))))

define define_av_ranges
$(1)_AV_RANGES             += $(call ysince,$(4),$(3))-$(4)
$(1)_AV_$(2)_RANGES        += $(call ysince,$(4),$(3))-$(4)
$(1)_AV_$(2)_$(3)yr_RANGES += $(call ysince,$(4),$(3))-$(4)
endef
$(foreach comp, $(COMPONENTS), \
$(foreach source, $($(comp)_AV_SOURCES), \
$(foreach intvl, $($(comp)_AV_$(source)_INTERVALS), \
$(foreach year, $(end$(intvl)), \
$(eval $(call define_av_ranges,$(comp),$(source),$(intvl),$(year)))))))

# not using vpath for now...
# vpath %.cpio     $(ARCHROOT)/history
# vpath %.tar      $(ARCHROOT)/history
# vpath %.ls $(PTMPROOT)/history
what:
	@echo Running $(me) ...
	@echo ARCHROOT = $(ARCHROOT)
	@echo PTMPROOT = $(PTMPROOT)
	@echo WORKROOT = $(WORKROOT)
	dmls -l $(ARCHROOT)/history |grep .nc.cpio
which:
	@echo Command paths: $(shell which $(commands))
vars:
	@echo HISTORIES=$(HISTORIES)
	@echo HISTDATES=$(HISTDATES)
	@echo HISTYEARS=$(HISTYEARS)
	@echo CHUNKSTART=$(CHUNKSTART)
	@echo PROCYEARS=$(PROCYEARS)
ifdef $(GRIDSPEC_SOURCE)
	@echo GRIDSPEC= $(GRIDSPEC)
	@echo GRIDSPEC_SOURCE= $(GRIDSPEC_SOURCE)
	@echo GRIDSPEC_FILES= $(GRIDSPEC_FILES)
	@echo GRIDSPEC_ROOT= $(GRIDSPEC_ROOT)
	@echo GRIDSPEC_PATHS= $(GRIDSPEC_PATHS)
endif
define comp_vars
phony += $(1)_vars
$(1)_vars:
	@echo $(1)_TILES=$($(1)_TILES)
	@echo $(1)_ZEXT=$($(1)_ZEXT)
	@echo $(1)_AV_SOURCES=$($(1)_AV_SOURCES)
	@echo $(1)_TS_FREQS=$($(1)_TS_FREQS)
endef
$(foreach comp, $(COMPONENTS), $(eval $(call comp_vars,$(comp))))

setup-dirs := $(foreach root, $(ARCHROOT) $(PTMPROOT) $(WORKROOT), \
              $(foreach stem, pp history tmp, \
              $(root)/$(stem)))
#ignore errors in setup-dirs, might not have write permission
#TODO: two ARCHes? read from someone else ARCHHIST, write to your own ARCHPP?
$(setup-dirs):
	-$(MKDIR) $@
setup: $(setup-dirs)
#@if( $(shell pwd) != $(WORKDIR) )echo cd $(WORKDIR) and run all commands from there.
phony += setup what which vars grids gridspec

# if any of the checkpoint files exist, bring the script down gracefully
# add a user-specified checkpoint file here as well
ckpt_files := /home/gfdl/flags/fre.checkpoint.$(JOB_ID) \
              /home/gfdl/flags/fre.checkpoint.$(HOSTNAME) \
              /home/gfdl/flags/fre.checkpoint.all \
              $(WORKDIR)/checkpoint
checkpoint: $(ckpt_files)
	$(HSMPUT) $(WORKDIR)
phony += $(ckpt_files) checkpoint

ifdef $(GRIDSPEC_SOURCE)
#gridspec
secondary += $(GRIDSPEC_PATHS)
gridspec: $(GRIDSPEC_PATHS)
$(GRIDSPEC_PATHS): $(GRIDSPEC_SOURCE)
	$(HSMGET) -a $(dir $<) -w $(WORKROOT) $@
#subcomponents inherit mosaic of parent
define do_gridspec
ifdef $(1)_TILES
GRIDSPEC_PATHS += $(GRIDSPEC)/$(2)_mosaic.nc
secondary += $(GRIDSPEC)/$(2)_mosaic.nc
$(GRIDSPEC)/$(2)_mosaic.nc: $(GRIDSPEC)/$(1)_mosaic.nc
	$(LOCALCP) $$< $$@
endif
endef
$(foreach comp, $(COMPONENTS), \
$(foreach sub, $($(comp)_SUBCOMPONENTS), \
$(eval $(call do_gridspec,$(comp),$(sub)))))
endif

#generic commands section

#checksum a directory tree
secondary += %.sum
%.sum: %
	test -d && cd $< && $(CHECKSUM) `find . -type f` > $@ || \
	test -f $< cd $(@D) && $(CHECKSUM) $(notdir $<) > $@

#generic netCDF commands

#list of static variables in FMS output file without a horizontal domain
secondary += %.s01
%.s01: %.nc
	echo `$(NCVARS) -s01   $<` | tr ' ' ',' > $@
#list of static variables in FMS output file
secondary += %.s0123
%.s0123: %.nc
	echo `$(NCVARS) -s0123 $<` | tr ' ' ',' > $@
#list of static variables in FMS output file with a horizontal domain
secondary += %.s23
%.s23: %.nc
	echo `$(NCVARS) -s23   $<` | tr ' ' ',' > $@
#list of variables in FMS output file with a horizontal domain
secondary += %.st23
%.st23: %.nc
	echo `$(NCVARS) -st23  $<` | tr ' ' ',' > $@

# plevel section
secondary += $(WORKDIR)/%.nc.plevel
# lorestime += $(WORKDIR)/%.plevel.nc
$(WORKDIR)/%.nc.plevel: $(WORKDIR)/%.nc
#we currently don't get the levels, use default ncep
	$(PLEVEL) -a -i $< -o $@

# fregrid section
# generalize to get variable list
define do_fregrid
ifdef $(1)_TILES
secondary += %.$(1).nc
%.$(1).nc: %.$(1)$(firstword $($(1)_TILES)).st23 $(foreach t,$($(1)_TILES),%.$(1)$(t).nc) $(GRIDSPEC_PATHS)
	$(FREGRID) --input_mosaic $(GRIDSPEC)/$(1)_mosaic.nc --input_file $$@ --scalar_field `cat $$<`
#correct grid information for backward compatibility
	$(NCRENAME) -d grid_xt,lon -d grid_yt,lat -v grid_xt,lon -v grid_yt,lat $$@
	$(NCATTED)  -a long_name,lon,m,c,"longitude" -a long_name,lat,m,c,"latitude" $$@
endif
endef
$(foreach comp, $(COMPONENTS), \
$(eval $(call do_fregrid,$(comp))))

#month 00 of year1 is needed for DJF
#this is the same as month 12 by frepp conventions
secondary += $(HISTSTART)00.%
$(HISTSTART)00.%: $(HISTSTART)12.%
	$(LOCALCP) $< $@

### subcomponents
define do_subcomponent
secondary += %.$($(2)_TS_$(3)_SOURCE)$(4).nc
%.$($(2)_TS_$(3)_SOURCE)$(4).nc: %.$($(1)_TS_$(3)_SOURCE)$(4).nc
	$(LOCALCP) $$< $$@
endef
$(foreach comp, $(COMPONENTS), \
$(foreach sub, $($(comp)_SUBCOMPONENTS), \
$(foreach freq, $($(sub)_TS_FREQS), \
$(foreach tile, $($(sub)_TILES), \
$(eval $(call do_subcomponent,$(comp),$(sub),$(freq),$(tile)))))))

### the static file
define do_static
secondary += $(PTMPROOT)/pp/$(1)/$(1).static.nc $(WORKROOT)/pp/$(1)/$(1).static.nc
secondary += $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE)$(firstword $($(1)_TILES)).s01 \
             $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE)$(firstword $($(1)_TILES)).nc  \
             $(WORKDIR)/$(PROCSTART)01.monthly.$(1).s23 \
             $(WORKDIR)/$(PROCSTART)01.monthly.$(1).nc
secondary += $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE).s0123 $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE).nc
$(1)_targets += $(PTMPROOT)/pp/$(1)/$(1).static.nc
lorestime += $(PTMPROOT)/pp/$(1)/$(1).static.nc $(WORKROOT)/pp/$(1)/$(1).static.nc
$(PTMPROOT)/pp/$(1)/$(1).static.nc: $(WORKROOT)/pp/$(1)/$(1).static.nc
	$(HSMPUT) $$^
ifdef $(1)_TILES
$(WORKROOT)/pp/$(1)/$(1).static.nc: $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE)$(firstword $($(1)_TILES)).s01 $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE)$(firstword $($(1)_TILES)).nc $(WORKDIR)/$(PROCSTART)01.monthly.$(1).s23 $(WORKDIR)/$(PROCSTART)01.monthly.$(1).nc
	@$(MKDIR) $$(@D)
	-$(NCSPLIT) -f $$@ -v `cat $$(word 1,$$^)` $$(word 2,$$^)
	-$(NCSPLIT) -f $$@ -v `cat $$(word 3,$$^)` $$(word 4,$$^)
else
$(WORKROOT)/pp/$(1)/$(1).static.nc: $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE).s0123 $(WORKDIR)/$(PROCSTART)01.$($(1)_TS_monthly_SOURCE).nc
	@$(MKDIR) $$(@D)
	-$(NCSPLIT) -f $$@ -v `cat $$(word 1,$$^)` $$(word 2,$$^)
endif
endef
$(foreach comp, $(COMPONENTS), \
$(eval $(call do_static,$(comp))))

#make intervals not in the original data: year|season
#TODO annual|seasonal from monthly: generalize annual for startmonth!=Jan
#TODO generalize year for different starting month
define do_year_and_seasons
secondary += $(WORKDIR)/$(2).$(3).$(1).nc $(WORKDIR)/$(2).$(3).$(1)%.nc
ifdef $(1)_TILES
#the % is the tiles, I could enter that directly
$(WORKDIR)/$(2).$(3).$(1)%.nc: $(foreach m, $($(3)_months), $(WORKDIR)/$(2)$(m).$($(1)_TS_monthly_SOURCE)%.nc$($(1)_ZEXT))
	$(TIMAVG) -o $$@ $$^
else
#this just repeats the above without the %
debug += $(WORKDIR)/$(2).$(3).$(1).nc:  $(foreach m, $($(3)_months), $(WORKDIR)/$(2)$(m).$($(1)_TS_monthly_SOURCE).nc$($(1)_ZEXT))
$(WORKDIR)/$(2).$(3).$(1).nc:  $(foreach m, $($(3)_months), $(WORKDIR)/$(2)$(m).$($(1)_TS_monthly_SOURCE).nc$($(1)_ZEXT))
	$(TIMAVG) -o $$@ $$^
endif
#the fully-processed snippet is saved in pp/comp/season/year.season.comp.nc
secondary += $(WORKROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
$(WORKROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc: $(WORKDIR)/$(2).$(3).$(1).nc
	@$(MKDIR) $$(@D)
	$(LOCALCP) $$< $$@
secondary     += $(PTMPROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
$(1)_targets  += $(PTMPROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
$(1)_snippets += $(PTMPROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
lorestime += $(PTMPROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc $(WORKROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
$(PTMPROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc: $(WORKROOT)/pp/$(1)/$(3)/$(2).$(3).$(1).nc
	$(HSMPUT) $$<
endef
$(foreach comp, $(COMPONENTS), \
$(foreach year, $(PROCYEARS), \
$(foreach season, annual $(SEASONS), \
$(eval $(call do_year_and_seasons,$(comp),$(year),$(season))))))

#intervals that exist in original data: month|day|sub-day
define do_months
ifneq ($(2),annual)
secondary += $(WORKDIR)/$(3)$(4).$(2).$(1)%.nc $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE)%.nc$($(1)_ZEXT)
# lorestime += $(WORKDIR)/$(3)$(4).$(2).$(1)%.nc $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE)%.nc$($(1)_ZEXT)
secondary += $(WORKDIR)/$(3)$(4).$(2).$(1).nc $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE).nc$($(1)_ZEXT)
# lorestime += $(WORKDIR)/$(3)$(4).$(2).$(1).nc $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE).nc$($(1)_ZEXT)
ifdef $(1)_TILES
$(WORKDIR)/$(3)$(4).$(2).$(1)%.nc: $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE)%.nc$($(1)_ZEXT)
	$(LOCALCP) $$< $$@
else
$(WORKDIR)/$(3)$(4).$(2).$(1).nc: $(WORKDIR)/$(3)$(4).$($(1)_TS_$(2)_SOURCE).nc$($(1)_ZEXT)
	$(LOCALCP) $$< $$@
endif
#the fully-processed snippet is saved in pp/comp/freq/year.month.freq.comp.nc
secondary += $(WORKROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
$(WORKROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc: $(WORKDIR)/$(3)$(4).$(2).$(1).nc
	@$(MKDIR) $$(@D)
	$(LOCALCP) $$< $$@
secondary     += $(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
$(1)_targets  += $(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
$(1)_snippets += $(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
lorestime += $(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc $(WORKROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
$(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc: $(WORKROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
	$(HSMPUT) $$<
ifeq ($(2),monthly)
secondary     += $(PTMPROOT)/pp/$(1)/$(4)/$(3).$(4).$(1).nc
$(1)_targets  += $(PTMPROOT)/pp/$(1)/$(4)/$(3).$(4).$(1).nc
$(1)_snippets += $(PTMPROOT)/pp/$(1)/$(4)/$(3).$(4).$(1).nc
$(PTMPROOT)/pp/$(1)/$(4)/$(3).$(4).$(1).nc: $(PTMPROOT)/pp/$(1)/$(2)/$(3)$(4).$(2).$(1).nc
	$(MKDIR) $$(@D) $$(call wpath,$$(@D))
	$(HSMGET) $$(call rpath,$$<)
	$(LOCALCP) $$< $$@
	$(LOCALCP) $$(call wpath,$$<) $$(call wpath,$$@)
endif
endif
endef
$(foreach comp, $(COMPONENTS), \
$(foreach freq, $($(comp)_TS_FREQS), \
$(foreach year, $(PROCYEARS), \
$(foreach month, $(MONTHS), \
$(eval $(call do_months,$(comp),$(freq),$(year),$(month)))))))

#time series section
#output pattern is comp.range.var.nc
define do_time_series
secondary += $(WORKDIR)/$(4).$(2).$(1).nc $(WORKROOT)/pp/$(1)/ts/$(2)/$(3)/$(4)
$(1)_targets += $(PTMPROOT)/pp/$(1)/ts/$(2)/$(3)/$(4)
lorestime += $(PTMPROOT)/pp/$(1)/ts/$(2)/$(3)/$(4) $(WORKROOT)/pp/$(1)/ts/$(2)/$(3)/$(4)
$(PTMPROOT)/pp/$(1)/ts/$(2)/$(3)/$(4): $(WORKROOT)/pp/$(1)/ts/$(2)/$(3)/$(4)
	$(HSMPUT) $$<
	touch $$@
$(WORKROOT)/pp/$(1)/ts/$(2)/$(3)/$(4): $(WORKDIR)/$(4).$(2).$(1).nc
	$(MKDIR) $$@
	$(NCSPLIT) -o $$@ $($(1)_TS_$(2)_VARS) $$<
#rename var.nc as comp.range.var.nc
	cd $$@ && ls | xargs -r -t -I foo mv foo $(1).$(4).foo
ifneq ($(2),annual)
$(WORKDIR)/$(4).$(2).$(1).nc: $(foreach m,$(shell $(LIST_MONTHS) $(4)),$(PTMPROOT)/pp/$(1)/$(2)/$(m).$(2).$(1).nc)
	$(HSMGET) $$(call rpath,$$^)
	$(NCRCAT) $$(call wpath,$$^) $$@
endif
endef
$(foreach comp, $(COMPONENTS), \
$(foreach freq, $($(comp)_TS_FREQS), \
$(foreach chunk, $($(comp)_TS_$(freq)_CHUNKS), \
$(foreach range, $($(comp)_TS_$(freq)_$(chunk)yr_RANGES), \
$(eval $(call do_time_series,$(comp),$(freq),$(chunk)yr,$(range)))))))

define do_annual_time_series
secondary += $(WORKDIR)/$(call ysince,$(3),$(2))-$(3).annual.$(1).nc
$(WORKDIR)/$(call ysince,$(3),$(2))-$(3).annual.$(1).nc: $(foreach y, $(call ylist,$(call ysince,$(3),$(2)),$(3)), $(PTMPROOT)/pp/$(1)/annual/$(y).annual.$(1).nc)
	$(HSMGET) $$(call rpath,$$^)
	$(NCRCAT) $$(call wpath,$$^) $$@
endef
$(foreach comp, $(COMPONENTS), \
$(foreach chunk, $($(comp)_TS_annual_CHUNKS), \
$(foreach year, $(end$(chunk)), \
$(eval $(call do_annual_time_series,$(comp),$(chunk),$(year))))))

##### time average section

define do_time_averages
#target := pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc
secondary += $(WORKROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc
#debug += $(WORKROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc: $(foreach y, $(call ylist,$(call ysince,$(4),$(3)),$(4)), $(PTMPROOT)/pp/$(1)/av/$(2)_1yr/$(1).$(y).$(5).nc)
$(WORKROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc: $(foreach y, $(call ylist,$(call ysince,$(4),$(3)),$(4)), $(PTMPROOT)/pp/$(1)/$(5)/$(y).$(5).$(1).nc)
	@$(MKDIR) $$(@D)
	$(HSMGET) $$(call rpath,$$^)
	$(NCRCAT) $$(call wpath,$$^) $$@
$(1)_targets += $(PTMPROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc
lorestime += $(PTMPROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc $(WORKROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc
$(PTMPROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc: $(WORKROOT)/pp/$(1)/av/$(2)_$(3)yr/$(1).$(call ysince,$(4),$(3))-$(4).$(5).nc
	$(HSMPUT) $$<
endef
$(foreach comp, $(COMPONENTS), \
$(foreach freq, $($(comp)_AV_FREQS), \
$(foreach chunk, $($(comp)_AV_$(freq)_CHUNKS), \
$(foreach year, $(end$(chunk)), \
$(foreach source, $($(freq)_SOURCES), \
$(eval $(call do_time_averages,$(comp),$(freq),$(chunk),$(year),$(source))))))))

#### data movement section

#get from archive
secondary += $(PTMPROOT)/history/%
$(PTMPROOT)/history/%: $(ARCHROOT)/history/%.nc.cpio
	$(HSMGET) $(call rpath,$@)/\*

#split history files by year: skip month 00 of first year of run
define list_hist
$(1)_container := $(PTMPROOT)/history/$(call container,$(1))
endef
$(foreach year, $(PROCYEARS), \
$(foreach month, 00 $(MONTHS), \
$(eval $(call list_hist,$(year)$(month)))))

define split_hist
ifneq ($(1),$(HISTSTART)00)
# 1=yyyymm
secondary += $(WORKDIR)/$(1).$(2).%.nc $(WORKDIR)/$(1).$(2).nc
$(WORKDIR)/$(1).$(2).%.nc: $($(1)_container)
	$(HSMGET) $$(call rpath,$$<)/\*.$(2).$$*.nc
	cd $(WORKDIR) && $(SPLITMONTH) `$(CONTAINER) $(1) $$(call wpath,$$<)/*.$(2).$$*.nc`
$(WORKDIR)/$(1).$(2).nc: $($(1)_container)
	$(HSMGET) $$(call rpath,$$<)/\*.$(2).nc
	cd $(WORKDIR) && $(SPLITMONTH) `$(CONTAINER) $(1) $$(call wpath,$$<)/*.$(2).nc`
endif
endef
$(foreach year, $(PROCYEARS), \
$(foreach month, 00 $(MONTHS), \
$(foreach comp, $(HISTORY_COMPONENTS), \
$(foreach freq, $($(comp)_TS_FREQS), \
$(eval $(call split_hist,$(year)$(month),$($(comp)_TS_$(freq)_SOURCE)))))))

#final targets
define do_component
phony += $(1) $(1)_snippets debug_$(1)_targets
$(1): $($(1)_targets)
	@echo Completed $$^
$(1)_snippets: $($(1)_snippets)
	@echo Completed $$^
debug_$(1)_targets:
	@echo $(1)_targets: $($(1)_targets)
endef
$(foreach comp, $(COMPONENTS), $(eval $(call do_component,$(comp))))
all: $(COMPONENTS)
	@echo DONE.

phony += debug
debug:
	@echo $(debug)

#the dot variables
.PRECIOUS:  $(secondary)
.SECONDARY: $(secondary) $(WORKDIR)/%
.LOW_RESOLUTION_TIME: $(lorestime)
.PHONY: $(phony)
.SUFFIXES:
.SUFFIXES: .cpio .nc .s01 .s0123 .s23 .st23 .sum .tar
